@@ -35,18 +35,65 @@
 #define ssh_forward_cancel ssh_channel_cancel_forward
 #endif
 
+
+/* Allocate a new SSH channel. */
+SCM_DEFINE_N (guile_ssh_make_channel, "%make-channel", 2, (SCM arg1, SCM flags),
+              "\
+Allocate a new SSH channel.\
+")
+#define FUNC_NAME s_guile_ssh_make_channel
+{
+    gssh_session_t *session_data = gssh_session_from_scm (arg1);
+    ssh_channel ch;
+
+    GSSH_VALIDATE_CONNECTED_SESSION (session_data, arg1, SCM_ARG1);
+    SCM_ASSERT (scm_is_integer (flags), flags, SCM_ARG2, FUNC_NAME);
+
+    ch = ssh_channel_new (session_data->ssh_session);
+
+    if (! ch)
+        return SCM_BOOL_F;
+
+    SCM channel = ssh_channel_to_scm (ch, arg1, scm_to_long (flags));
+    gssh_session_add_channel_x (session_data, channel);
+
+    return channel;
+}
+#undef FUNC_NAME
+
+SCM_DEFINE_1 (guile_ssh_is_channel_p, "channel?", (SCM x),
+              "\
+Return #t if X is a SSH channel, #f otherwise.\
+")
+{
+#if USING_GUILE_BEFORE_2_2
+    return scm_from_bool (SCM_SMOB_PREDICATE (channel_tag, x));
+#else
+    return scm_from_bool (SCM_PORTP (x)
+                          && SCM_PORT_TYPE (x) == channel_tag);
+#endif
+}
+
+#if USING_GUILE_BEFORE_2_2
+SCM
+equalp_channel (SCM x1, SCM x2)
+{
+    return compare_objects(x1, x2, gssh_channel_from_scm);
+}
+#endif
+
 
 /* Procedures */
 
-SCM_DEFINE (guile_ssh_channel_open_session, "channel-open-session", 1, 0, 0,
-            (SCM channel),
+SCM_DEFINE_1 (guile_ssh_channel_open_session, "channel-open-session",
+              (SCM channel),
             "\
 Open a new session and mark the channel CHANNEL as opened port.\n\
 Return value is undefined.\
 ")
 #define FUNC_NAME s_guile_ssh_channel_open_session
 {
-  struct channel_data *data = _scm_to_channel_data (channel);
+  gssh_channel_t *data = gssh_channel_from_scm (channel);
   int res;
   GSSH_VALIDATE_CHANNEL_DATA (data, channel, FUNC_NAME);
 
@@ -68,14 +115,14 @@ Return value is undefined.\
 #undef FUNC_NAME
 
 /* Run a shell command CMD without an interactive shell. */
-SCM_DEFINE (guile_ssh_channel_request_exec, "channel-request-exec", 2, 0, 0,
-            (SCM channel, SCM cmd),
+SCM_DEFINE_N (guile_ssh_channel_request_exec, "channel-request-exec", 2,
+              (SCM channel, SCM cmd),
             "\
 Run a shell command CMD without an interactive shell.\
 ")
 #define FUNC_NAME s_guile_ssh_channel_request_exec
 {
-  struct channel_data *data = _scm_to_channel_data (channel);
+  gssh_channel_t *data = gssh_channel_from_scm (channel);
   int res;
   char *c_cmd;                  /* Command to execute. */
 
@@ -100,22 +147,21 @@ Run a shell command CMD without an interactive shell.\
 }
 #undef FUNC_NAME
 
-SCM_DEFINE (guile_ssh_channel_get_exit_status,
-            "channel-get-exit-status", 1, 0, 0,
-            (SCM channel),
-            "\
+SCM_DEFINE_1 (guile_ssh_channel_get_exit_status, "channel-get-exit-status",
+              (SCM channel),
+              "\
 Get the exit status of the channel (error code from the executed \
 instruction).  Return the exist status, or #f if no exit status has been \
 returned (yet). \
 ")
 #define FUNC_NAME s_guile_ssh_channel_get_exit_status
 {
-  struct channel_data *cd = NULL;
+  gssh_channel_t *cd = NULL;
   int res;
 
   GSSH_VALIDATE_OPEN_CHANNEL (channel, SCM_ARG1, FUNC_NAME);
 
-  cd = _scm_to_channel_data (channel);
+  cd = gssh_channel_from_scm (channel);
 
   if (! _gssh_channel_parent_session_connected_p (cd))
     guile_ssh_error1 (FUNC_NAME, "Parent session is not connected", channel);
@@ -131,17 +177,17 @@ returned (yet). \
 }
 #undef FUNC_NAME
 
-SCM_DEFINE (guile_ssh_channel_request_send_exit_status,
-            "channel-request-send-exit-status", 2, 0, 0,
-            (SCM channel, SCM exit_status),
-            "\
+SCM_DEFINE_N (guile_ssh_channel_request_send_exit_status,
+              "channel-request-send-exit-status", 2,
+              (SCM channel, SCM exit_status),
+              "\
 Send the exit status to the remote process (as described in RFC 4254, section\n\
 6.10).\n\
 Return value is undefined.\
 ")
 #define FUNC_NAME s_guile_ssh_channel_request_send_exit_status
 {
-  struct channel_data *cd = _scm_to_channel_data (channel);
+  gssh_channel_t *cd = gssh_channel_from_scm (channel);
   int res;
 
   GSSH_VALIDATE_OPEN_CHANNEL (channel, SCM_ARG1, FUNC_NAME);
@@ -165,15 +211,15 @@ Return value is undefined.\
 }
 #undef FUNC_NAME
 
-SCM_DEFINE (guile_ssh_channel_request_pty, "channel-request-pty", 1, 0, 0,
-            (SCM channel),
-            "\
+SCM_DEFINE_1 (guile_ssh_channel_request_pty, "channel-request-pty",
+              (SCM channel),
+              "\
 Request a PTY (pseudo terminal).\n\
 Return value is undefined.\
 ")
 #define FUNC_NAME s_guile_ssh_channel_request_pty
 {
-  struct channel_data *data = _scm_to_channel_data (channel);
+  gssh_channel_t *data = gssh_channel_from_scm (channel);
   int res;
 
   GSSH_VALIDATE_OPEN_CHANNEL (channel, SCM_ARG1, FUNC_NAME);
@@ -193,15 +239,15 @@ Return value is undefined.\
 }
 #undef FUNC_NAME
 
-SCM_DEFINE (guile_ssh_channel_request_shell, "channel-request-shell", 1, 0, 0,
-            (SCM channel),
-            "\
+SCM_DEFINE_1 (guile_ssh_channel_request_shell, "channel-request-shell",
+              (SCM channel),
+              "\
 Request a shell.\n\
 Return value is undefined.\
 ")
 #define FUNC_NAME s_guile_ssh_channel_request_shell
 {
-  struct channel_data *data = _scm_to_channel_data (channel);
+  gssh_channel_t *data = gssh_channel_from_scm (channel);
   int res;
 
   GSSH_VALIDATE_OPEN_CHANNEL (channel, SCM_ARG1, FUNC_NAME);
@@ -223,15 +269,15 @@ Return value is undefined.\
 
 /* Set an environment variable NAME to value VALUE
    Return value is undefined. */
-SCM_DEFINE (guile_ssh_channel_request_env, "channel-request-env", 3, 0, 0,
-            (SCM channel, SCM name, SCM value),
-            "\
+SCM_DEFINE_N (guile_ssh_channel_request_env, "channel-request-env", 3,
+              (SCM channel, SCM name, SCM value),
+              "\
 Set an environment variable NAME to value VALUE.\n\
 Return value is undefined.\
 ")
 #define FUNC_NAME s_guile_ssh_channel_request_env
 {
-  struct channel_data *data = _scm_to_channel_data (channel);
+  gssh_channel_t *data = gssh_channel_from_scm (channel);
   char *c_name;
   char *c_value;
   int res;
@@ -285,16 +331,16 @@ _ssh_result_to_symbol (int res)
 }
 #undef FUNC_NAME
 
-SCM_GSSH_DEFINE (guile_ssh_channel_open_forward,
-                 "%channel-open-forward", 5,
-                 (SCM channel, SCM remote_host, SCM remote_port,
-                  SCM source_host, SCM local_port))
+SCM_DEFINE_N (guile_ssh_channel_open_forward, "%channel-open-forward", 5,
+              (SCM channel, SCM remote_host, SCM remote_port,
+               SCM source_host, SCM local_port),
+              "")
 #define FUNC_NAME s_guile_ssh_channel_open_forward
 {
-  struct channel_data *cd = _scm_to_channel_data (channel);
+  gssh_channel_t *cd = gssh_channel_from_scm (channel);
   char *c_remote_host = NULL;
   char *c_source_host = NULL;
-  struct session_data *sd = NULL;
+  gssh_session_t *sd = NULL;
   int res;
 
   SCM_ASSERT (scm_is_string (remote_host), remote_host, SCM_ARG2, FUNC_NAME);
@@ -308,7 +354,7 @@ SCM_GSSH_DEFINE (guile_ssh_channel_open_forward,
   if (! _gssh_channel_parent_session_connected_p (cd))
     guile_ssh_error1 (FUNC_NAME, "Parent session is not connected", channel);
 
-  sd = _scm_to_session_data (cd->session);
+  sd = gssh_session_from_scm (cd->session);
   if (! sd)
     guile_ssh_error1 (FUNC_NAME, "Session is freed: ", cd->session);
 
@@ -349,12 +395,12 @@ SCM_GSSH_DEFINE (guile_ssh_channel_open_forward,
 }
 #undef FUNC_NAME
 
-SCM_GSSH_DEFINE (guile_ssh_channel_listen_forward,
-                 "%channel-listen-forward", 3,
-                 (SCM session, SCM address, SCM port))
+SCM_DEFINE_N (guile_ssh_channel_listen_forward, "%channel-listen-forward", 3,
+              (SCM session, SCM address, SCM port),
+              "")
 #define FUNC_NAME s_guile_ssh_channel_listen_forward
 {
-  struct session_data *sd = _scm_to_session_data (session);
+  gssh_session_t *sd = gssh_session_from_scm (session);
   char *c_address = NULL;
   int bound_port;
   int res;
@@ -389,12 +435,12 @@ SCM_GSSH_DEFINE (guile_ssh_channel_listen_forward,
 }
 #undef FUNC_NAME
 
-SCM_GSSH_DEFINE (guile_ssh_channel_accept_forward,
-                 "%channel-accept-forward", 2,
-                 (SCM session, SCM timeout))
+SCM_DEFINE_N (guile_ssh_channel_accept_forward, "%channel-accept-forward", 2,
+              (SCM session, SCM timeout),
+              "")
 #define FUNC_NAME s_guile_ssh_channel_accept_forward
 {
-  struct session_data *sd = _scm_to_session_data (session);
+  gssh_session_t *sd = gssh_session_from_scm (session);
   ssh_channel c_channel = NULL;
   SCM channel = SCM_BOOL_F;
   int port;
@@ -406,8 +452,7 @@ SCM_GSSH_DEFINE (guile_ssh_channel_accept_forward,
                                           &port);
   if (c_channel)
     {
-      channel = _scm_from_channel_data (c_channel, session,
-                                        SCM_RDNG | SCM_WRTNG);
+      channel = ssh_channel_to_scm (c_channel, session, SCM_RDNG | SCM_WRTNG);
       SCM_SET_CELL_TYPE (channel, SCM_CELL_TYPE (channel) | SCM_OPN);
     }
 
@@ -416,12 +461,12 @@ SCM_GSSH_DEFINE (guile_ssh_channel_accept_forward,
 #undef FUNC_NAME
 
 /* FIXME: Should it be defined in some other module? */
-SCM_GSSH_DEFINE (guile_ssh_channel_cancel_forward,
-                 "channel-cancel-forward", 3,
-                 (SCM session, SCM address, SCM port))
+SCM_DEFINE_N (guile_ssh_channel_cancel_forward, "channel-cancel-forward", 3,
+              (SCM session, SCM address, SCM port),
+              "")
 #define FUNC_NAME s_guile_ssh_channel_cancel_forward
 {
-  struct session_data *sd = _scm_to_session_data (session);
+  gssh_session_t *sd = gssh_session_from_scm (session);
   char *c_address = NULL;
   int res;
 
@@ -442,16 +487,15 @@ SCM_GSSH_DEFINE (guile_ssh_channel_cancel_forward,
 }
 #undef FUNC_NAME
 
-SCM_DEFINE (guile_ssh_channel_set_pty_size_x,
-            "channel-set-pty-size!", 3, 0, 0,
-            (SCM channel, SCM col, SCM row),
-            "\
+SCM_DEFINE_N (guile_ssh_channel_set_pty_size_x, "channel-set-pty-size!", 3,
+              (SCM channel, SCM col, SCM row),
+              "\
 Change size of the PTY to columns COL and rows ROW.\n\
 eturn value is undefined.\
 ")
 #define FUNC_NAME s_guile_ssh_channel_set_pty_size_x
 {
-  struct channel_data *data = _scm_to_channel_data (channel);
+  gssh_channel_t *data = gssh_channel_from_scm (channel);
 
   GSSH_VALIDATE_OPEN_CHANNEL (channel, SCM_ARG1, FUNC_NAME);
   SCM_ASSERT (scm_is_unsigned_integer (col, 0, UINT32_MAX), col,
@@ -470,17 +514,16 @@ eturn value is undefined.\
 }
 #undef FUNC_NAME
 
-SCM_DEFINE (guile_ssh_channel_set_stream_x,
-            "channel-set-stream!", 2, 0, 0,
-            (SCM channel, SCM stream_name),
-            "\
+SCM_DEFINE_N (guile_ssh_channel_set_stream_x, "channel-set-stream!", 2,
+              (SCM channel, SCM stream_name),
+              "\
 Set stream STREAM_NAME for channel CHANNEL.  STREAM_NAME must be one of the \n\
 following symbols: \"stdout\" (default), \"stderr\".\n\
 Return value is undefined.\
 ")
 #define FUNC_NAME s_guile_ssh_channel_set_stream_x
 {
-  struct channel_data *cd = _scm_to_channel_data (channel);
+  gssh_channel_t *cd = gssh_channel_from_scm (channel);
 
   GSSH_VALIDATE_OPEN_CHANNEL (channel, SCM_ARG1, FUNC_NAME);
   SCM_ASSERT (scm_is_symbol (stream_name), stream_name, SCM_ARG2, FUNC_NAME);
@@ -507,16 +550,15 @@ Return value is undefined.\
 }
 #undef FUNC_NAME
 
-SCM_DEFINE (guile_ssh_channel_get_stream,
-            "channel-get-stream", 1, 0, 0,
-            (SCM channel),
-            "\
+SCM_DEFINE_1 (guile_ssh_channel_get_stream, "channel-get-stream", (SCM channel),
+              "\
 Get current stream name from CHANNEL.  Throw `guile-ssh-error' on error.\n\
 Return one of the following symbols: \"stdout\", \"stderr\".\
 ")
 #define FUNC_NAME s_guile_ssh_channel_get_stream
 {
-  struct channel_data *cd = _scm_to_channel_data (channel);
+  gssh_channel_t *cd = gssh_channel_from_scm (channel);
+  SCM result = SCM_UNDEFINED;
 
   GSSH_VALIDATE_OPEN_CHANNEL (channel, SCM_ARG1, FUNC_NAME);
 
@@ -524,27 +566,32 @@ Return one of the following symbols: \"stdout\", \"stderr\".\
     guile_ssh_error1 (FUNC_NAME, "Parent session is not connected", channel);
 
   if (cd->is_stderr == 0)
-    return scm_from_locale_symbol ("stdout");
+    {
+      result = scm_from_locale_symbol ("stdout");
+    }
   if (cd->is_stderr == 1)
-    return scm_from_locale_symbol ("stderr");
+    {
+      result = scm_from_locale_symbol ("stderr");
+    }
+  else
+    {
+      guile_ssh_error1 (FUNC_NAME, "Wrong stream.",
+                        scm_from_int (cd->is_stderr));
+    }
 
-  guile_ssh_error1 (FUNC_NAME, "Wrong stream.",
-                    scm_from_int (cd->is_stderr));
-
-  return SCM_UNDEFINED;
+  return result;
 }
 #undef FUNC_NAME
 
-SCM_DEFINE (guile_ssh_channel_get_session,
-            "channel-get-session", 1, 0, 0,
-            (SCM channel),
-            "\
+SCM_DEFINE_1 (guile_ssh_channel_get_session, "channel-get-session",
+              (SCM channel),
+              "\
 Get the session to which belongs the CHANNEL.  Throw `guile-ssh-error' on an \n\
 error.  Return the session.\
 ")
 #define FUNC_NAME s_guile_ssh_channel_get_session
 {
-  struct channel_data *cd = _scm_to_channel_data (channel);
+  gssh_channel_t *cd = gssh_channel_from_scm (channel);
   GSSH_VALIDATE_CHANNEL_DATA (cd, channel, FUNC_NAME);
   return cd->session;
 }
@@ -553,21 +600,28 @@ error.  Return the session.\
 
 /* Predicates */
 
-SCM_DEFINE (guile_ssh_channel_is_open_p, "channel-open?", 1, 0, 0,
-            (SCM channel),
-            "Return #t if channel CHANNEL is open, #f otherwise.")
+SCM_DEFINE_1 (guile_ssh_channel_is_open_p, "channel-open?", (SCM channel),
+              "Return #t if channel CHANNEL is open, #f otherwise.")
 {
-  struct channel_data *data = _scm_to_channel_data (channel);
-  if (data && ssh_channel_is_open (data->ssh_channel))
+  gssh_channel_t *data = gssh_channel_from_scm (channel);
+
+  if (data == NULL)
+      return SCM_BOOL_F;
+
+  if (data->is_remote_closed)
+      return SCM_BOOL_F;
+
+  if (ssh_channel_is_open (data->ssh_channel))
     return SCM_BOOL_T;
+
   return SCM_BOOL_F;
 }
 
-SCM_GSSH_DEFINE (gssh_channel_send_eof, "%channel-send-eof",
-                 1, (SCM channel))
+SCM_DEFINE_1 (gssh_channel_send_eof, "%channel-send-eof", (SCM channel),
+              "")
 #define FUNC_NAME s_gssh_channel_send_eof
 {
-  struct channel_data *cd = _scm_to_channel_data (channel);
+  gssh_channel_t *cd = gssh_channel_from_scm (channel);
   scm_t_bits pt_bits;
   int rc;
 
@@ -593,15 +647,14 @@ SCM_GSSH_DEFINE (gssh_channel_send_eof, "%channel-send-eof",
 }
 #undef FUNC_NAME
 
-SCM_DEFINE (guile_ssh_channel_is_eof_p, "channel-eof?", 1, 0, 0,
-            (SCM channel),
-            "\
+SCM_DEFINE_1 (guile_ssh_channel_is_eof_p, "channel-eof?", (SCM channel),
+              "\
 Return #t if remote has set EOF, #f otherwise.\n\
 Throw `guile-ssh-error' if the channel has been closed and freed.\
 ")
 #define FUNC_NAME s_guile_ssh_channel_is_eof_p
 {
-  struct channel_data *data = _scm_to_channel_data (channel);
+  gssh_channel_t *data = gssh_channel_from_scm (channel);
   GSSH_VALIDATE_CHANNEL_DATA (data, channel, FUNC_NAME);
   return scm_from_bool (ssh_channel_is_eof (data->ssh_channel));
 }
